@@ -1,10 +1,20 @@
+import re
+from collections import defaultdict
+from ctypes import c_float
 from sys import stderr
-from typing import Union, Iterable, Tuple, Dict, Callable, Any
+from typing import Any, Callable, Dict, Iterable, Tuple, Union
 
-from .errors import MoreThanOneAltAllele
+from .errors import MoreThanOneAltAllele, NotExactlyOneValue
 
 
-class NoValue:
+def float32(val: str) -> float:
+    return c_float(float(val)).value
+
+
+# If NoValue inherits from str, re.search("something", NoValue()) does not error
+# but just comes up empty handed, which is convenient behaviour.
+# This way, we do not have to special case / monkey patch / wrap the regex module.
+class NoValue(str):
     def __lt__(self, other):
         return False
 
@@ -20,9 +30,23 @@ class NoValue:
     def __eq__(self, other):
         return False
 
+    def __ne__(self, other):
+        return True
+
     def __bool__(self):
         return False
 
+    def __str__(self):
+        # nonexistent fields will result in an empty string
+        # should be configurable in upcoming versions
+        return ""
+
+    def __repr__(self):
+        return f"{self.__class__.__name__}()"
+
+    def __hash__(self) -> int:
+        return super().__hash__()
+
 
 NA = NoValue()
 
@@ -41,27 +65,63 @@ class InfoTuple:
             return NA
         return values
 
+    def __str__(self):
+        return self.values.__str__()
+
+    def __repr__(self):
+        return f"{self.__class__.__name__}(values={self.values!r})"
+
+    def __len__(self):
+        return len(self.values)
+
+    def __eq__(self, other):
+        value_type = type(self.values)
+        if isinstance(other, InfoTuple) and isinstance(other.values, value_type):
+            return self.values == other.values
+        elif isinstance(other, value_type):
+            return self.values == other
+        else:
+            raise TypeError(
+                f"Incomparable: "
+                f"{type(self)} (value type: {type(self.values)}), "
+                f"{type(other)}"
+                + (
+                    f" (value type: {type(other.values)})"
+                    if isinstance(other, InfoTuple)
+                    else ""
+                )
+            )
+
+    def __hash__(self):
+        return hash(self.values)
+
 
 IntFloatStr = Union[int, float, str]
 
 
-def type_info(value, number="."):
+def type_info(value, number=".", field=None, record_idx=None):
     if value is None:
         return NA
+    if number == "0":
+        return NA if value is None else value
+    if number == "A":
+        if len(value) != 1:
+            raise MoreThanOneAltAllele()
+        return value[0] if value[0] is not None else NA
+    if number == "R":
+        if len(value) != 2:
+            raise MoreThanOneAltAllele()
+        return InfoTuple(value)
+    if number == "1":
+        if isinstance(value, tuple):
+            if len(value) != 1:
+                raise NotExactlyOneValue(field, len(value), record_idx)
+            return value[0] if value[0] is not None else NA
+        return value if value is not None else NA
     if isinstance(value, tuple):
-        if number not in {"A", "R"}:
-            return InfoTuple(value)
-        else:
-            if number == "A":
-                if len(value) != 1:
-                    raise MoreThanOneAltAllele()
-                value = value[0]
-                return value if value is not None else NA
-            if number == "R":
-                if len(value) != 2:
-                    raise MoreThanOneAltAllele()
-                return InfoTuple(value)
-    return value
+        return InfoTuple(value)
+    else:
+        return InfoTuple((value,))
 
 
 class PosRange:
@@ -96,7 +156,7 @@ class PosRange:
         return f"(start: {self.start}, end: {self.end}, length: {self.length})"
 
     def __repr__(self):
-        return self.__str__()
+        return f"{self.__class__.__name__}(start={self.start!r}, end={self.end!r})"
 
 
 def na_func() -> NoValue:
@@ -132,15 +192,127 @@ class AnnotationEntry:
     def description(self):
         return self._description
 
+    def __repr__(self):
+        return (
+            f"{self.__class__.__name__}("
+            f"name={self._name!r}, "
+            f"typefunc={self._typefunc!r}, "
+            f"nafunc={self._nafunc!r}, "
+            f"description={self._description!r}"
+            f")"
+        )
+
 
 class AnnotationListEntry(AnnotationEntry):
     def __init__(
-        self, name: str, sep: str, typefunc: Callable[[str], Any] = None, **kwargs
+        self,
+        name: str,
+        sep: str,
+        typefunc: Callable[[str], Any] = None,
+        inner_typefunc: Callable[[str], Any] = lambda x: x,
+        **kwargs,
     ):
-        typefunc = typefunc if typefunc else lambda v: [x.strip() for x in v.split(sep)]
+        typefunc = (
+            typefunc
+            if typefunc
+            else lambda v: [inner_typefunc(x.strip()) for x in v.split(sep)]
+        )
         super().__init__(name, typefunc, nafunc=lambda: [], **kwargs)
 
 
+class AnnotationListDictEntry(AnnotationEntry):
+    def __init__(self, name: str, nafunc=lambda: None, **kwargs):
+        def typefunc(x):
+            key_value_tuples = map(lambda v: v.split(":", maxsplit=1), x.split("&"))
+            mapping = defaultdict(list)
+            for key, value in key_value_tuples:
+                mapping[key].append(value)
+            # make sure non-existent keys do not return list() on __getitem__ calls
+            # in the filter expression (without actually building a new dict from
+            # the defaultdict)
+            mapping.default_factory = nafunc
+            return mapping
+
+        super().__init__(name=name, typefunc=typefunc, **kwargs)
+
+
+class RangeTotal(object):
+    def __init__(self, r, total):
+        self.range = r
+        self.total = total
+
+    @classmethod
+    def from_vep_string(cls, value: str) -> "RangeTotal":
+        v = value.split("/")
+        r = [int(s) for s in v[0].split("-")]
+        if len(r) == 1:
+            return cls(range(r[0], r[0] + 1), int(v[1]))
+        elif len(r) == 2:
+            return cls(range(r[0], r[1] + 1), int(v[1]))
+        else:
+            raise ValueError(
+                "Found more than two values separated by '-', "
+                "expected only a single int, or two ints separated by '-'."
+            )
+
+    def __str__(self):
+        if len(self.range) == 1:
+            return f"number / total: {self.range.start} / {self.total}"
+        else:
+            return (
+                f"range / total: "
+                f"{self.range.start} - {self.range.stop - 1} / {self.total}"
+            )
+
+    def __repr__(self):
+        return (
+            f"{self.__class__.__name__}"
+            f"(range=range({self.range.start!r}, {self.range.stop!r}), "
+            f"total={self.total!r})"
+        )
+
+
+class AnnotationRangeTotalEntry(AnnotationEntry):
+    def __init__(self, name: str, **kwargs):
+        super().__init__(name, typefunc=RangeTotal.from_vep_string, **kwargs)
+
+
+class NumberTotal(object):
+    def __init__(self, number, total):
+        self.number = number
+        self.total = total
+
+    @classmethod
+    def from_vep_string(cls, value: str) -> "NumberTotal":
+        v = value.split("/")
+        return cls(int(v[0]), int(v[1]))
+
+    def __str__(self):
+        return f"number / total: {self.number} / {self.total}"
+
+    def __repr__(self):
+        return (
+            f"{self.__class__.__name__}(number={self.number!r}, total={self.total!r})"
+        )
+
+
+class AnnotationNumberTotalEntry(AnnotationEntry):
+    def __init__(self, name: str, **kwargs):
+        super().__init__(name, typefunc=NumberTotal.from_vep_string, **kwargs)
+
+
+PREDICTION_SCORE_REGEXP: re.Pattern = re.compile(r"(.*)\((.*)\)")
+
+
+class AnnotationPredictionScoreEntry(AnnotationEntry):
+    def __init__(self, name: str, **kwargs):
+        def typefunc(x: str) -> Dict[str, float]:
+            match = PREDICTION_SCORE_REGEXP.findall(x)[0]
+            return {match[0]: float32(match[1])}
+
+        super().__init__(name, typefunc=typefunc, nafunc=lambda: dict(), **kwargs)
+
+
 class DefaultAnnotationEntry(AnnotationEntry):
     def __init__(self, name: str):
         super().__init__(name)
@@ -150,7 +322,7 @@ class AnnotationTyper:
     def __init__(self, mapping: Dict[str, AnnotationEntry]):
         self._mapping = mapping
 
-    def get_entry(self, key: str) -> Tuple[str, AnnotationType]:
+    def get_entry(self, key: str) -> AnnotationEntry:
         entry = self._mapping.get(key)
         if not entry:
             print(
@@ -185,7 +357,10 @@ KNOWN_ANN_TYPE_MAP_SNPEFF = {
     "CDS.pos / CDS.length": AnnotationEntry("CDS", PosRange.from_snpeff_str),
     "AA.pos / AA.length": AnnotationEntry("AA", PosRange.from_snpeff_str),
     "Distance": AnnotationEntry("Distance", str),
-    "ERRORS / WARNINGS / INFO": AnnotationListEntry("ERRORS / WARNINGS / INFO", "/",),
+    "ERRORS / WARNINGS / INFO": AnnotationListEntry(
+        "ERRORS / WARNINGS / INFO",
+        "/",
+    ),
 }
 
 # fields, types and description taken from:
@@ -232,12 +407,12 @@ KNOWN_ANN_TYPE_MAP_VEP = {
     "ENSP": AnnotationEntry(
         "ENSP", description="The Ensembl protein identifier of the affected transcript"
     ),
-    # TODO parse flags: "cds_start_NF: CDS 5' incomplete, cds_end_NF: CDS 3' incomplete"
-    # "FLAGS": AnnotationEntry(
-    #     "FLAGS",
-    #     description="Transcript quality flags:
-    #     cds_start_NF: CDS 5' incomplete, cds_end_NF: CDS 3' incomplete",
-    # ),
+    "FLAGS": AnnotationListEntry(
+        "FLAGS",
+        sep="&",
+        description="Transcript quality flags: "
+        "cds_start_NF: CDS 5' incomplete, cds_end_NF: CDS 3' incomplete",
+    ),
     "SWISSPROT": AnnotationEntry(
         "SWISSPROT",
         description="Best match UniProtKB/Swiss-Prot accession of protein product",
@@ -259,18 +434,14 @@ KNOWN_ANN_TYPE_MAP_VEP = {
     ),
     # "NEAREST": AnnotationEntry("NEAREST",
     #               description="Identifier(s) of nearest transcription start site"),
-    # TODO custom type:
-    #  "the SIFT prediction and/or score, with both given as prediction(score)"
-    # "SIFT": AnnotationEntry(
-    #     "SIFT",
-    #     description="The SIFT prediction and/or score,"
-    #     " with both given as prediction(score)",
-    # ),
-    # TODO custom type:
-    #  "the PolyPhen prediction and/or score"
-    # "PolyPhen": AnnotationEntry(
-    #     "PolyPhen", description="The PolyPhen prediction and/or score"
-    # ),
+    "SIFT": AnnotationPredictionScoreEntry(
+        "SIFT",
+        description="The SIFT prediction and/or score,"
+        " with both given as prediction(score)",
+    ),
+    "PolyPhen": AnnotationPredictionScoreEntry(
+        "PolyPhen", description="The PolyPhen prediction and/or score"
+    ),
     "MOTIF_NAME": AnnotationEntry(
         "MOTIF_NAME",
         description="The source and identifier of a transcription factor binding "
@@ -281,17 +452,15 @@ KNOWN_ANN_TYPE_MAP_VEP = {
         int,
         description="The relative position of the variation in the aligned TFBP",
     ),
-    # TODO parse flag:
-    #  "a flag indicating if the variant falls in a high information position of a
-    #  transcription factor binding profile (TFBP)"
-    # "HIGH_INF_POS": AnnotationEntry(
-    #     "HIGH_INF_POS",
-    #     description="A flag indicating if the variant falls in a high information "
-    #                 "position of a transcription factor binding profile (TFBP)",
-    # ),
+    "HIGH_INF_POS": AnnotationEntry(
+        "HIGH_INF_POS",
+        typefunc="Y".__eq__,
+        description="A flag indicating if the variant falls in a high information "
+        "position of a transcription factor binding profile (TFBP)",
+    ),
     "MOTIF_SCORE_CHANGE": AnnotationEntry(
         "MOTIF_SCORE_CHANGE",
-        float,
+        float32,
         description="The difference in motif score of the reference and variant "
         "sequences for the TFBP",
     ),
@@ -300,29 +469,25 @@ KNOWN_ANN_TYPE_MAP_VEP = {
         ",",
         description="List of cell types and classifications for regulatory feature",
     ),
-    # TODO parse flag:
-    #  "a flag indicating if the transcript is denoted as the canonical transcript
-    #  for this gene"
-    # "CANONICAL": AnnotationEntry(
-    #     "CANONICAL",
-    #     description="A flag indicating if the transcript is denoted as the canonical "
-    #                 "transcript for this gene",
-    # ),
+    "CANONICAL": AnnotationEntry(
+        "CANONICAL",
+        typefunc="YES".__eq__,
+        description="A flag indicating if the transcript is denoted as the canonical "
+        "transcript for this gene",
+    ),
     "CCDS": AnnotationEntry(
         "CCDS", description="The CCDS identifer for this transcript, where applicable"
     ),
-    # TODO custom type: "the intron number (out of total number)"
-    # "INTRON": AnnotationEntry(
-    #     "INTRON", description="The intron number (out of total number)"
-    # ),
-    # TODO custom type: "the exon number (out of total number)"
-    # "EXON": AnnotationEntry(
-    #     "EXON", description="The exon number (out of total number)"
-    # ),
-    # "DOMAINS": AnnotationEntry(
-    #     "DOMAINS",
-    #     description="The source and identifer of any overlapping protein domains",
-    # ),
+    "INTRON": AnnotationNumberTotalEntry(
+        "INTRON", description="The intron number (out of total number)"
+    ),
+    "EXON": AnnotationRangeTotalEntry(
+        "EXON", description="The exon index range (out of total number of exons)"
+    ),
+    "DOMAINS": AnnotationListDictEntry(
+        "DOMAINS",
+        description="The source and identifer of any overlapping protein domains",
+    ),
     "DISTANCE": AnnotationEntry(
         "DISTANCE", int, description="Shortest distance from variant to transcript"
     ),
@@ -335,120 +500,121 @@ KNOWN_ANN_TYPE_MAP_VEP = {
     #     "FREQS", description="Frequencies of overlapping variants used in filtering"
     # ),
     "AF": AnnotationEntry(
-        "AF", float, description="Frequency of existing variant in 1000 Genomes"
+        "AF", float32, description="Frequency of existing variant in 1000 Genomes"
     ),
     "AFR_AF": AnnotationEntry(
         "AFR_AF",
-        float,
+        float32,
         description="Frequency of existing variant in 1000 Genomes combined "
         "African population",
     ),
     "AMR_AF": AnnotationEntry(
         "AMR_AF",
-        float,
+        float32,
         description="Frequency of existing variant in 1000 Genomes combined "
         "American population",
     ),
     "ASN_AF": AnnotationEntry(
         "ASN_AF",
-        float,
+        float32,
         description="Frequency of existing variant in 1000 Genomes combined "
         "Asian population",
     ),
     "EUR_AF": AnnotationEntry(
         "EUR_AF",
-        float,
+        float32,
         description="Frequency of existing variant in 1000 Genomes combined "
         "European population",
     ),
     "EAS_AF": AnnotationEntry(
         "EAS_AF",
-        float,
+        float32,
         description="Frequency of existing variant in 1000 Genomes combined "
         "East Asian population",
     ),
     "SAS_AF": AnnotationEntry(
         "SAS_AF",
-        float,
+        float32,
         description="Frequency of existing variant in 1000 Genomes combined "
         "South Asian population",
     ),
     "AA_AF": AnnotationEntry(
         "AA_AF",
-        float,
+        float32,
         description="Frequency of existing variant in NHLBI-ESP "
         "African American population",
     ),
     "EA_AF": AnnotationEntry(
         "EA_AF",
-        float,
+        float32,
         description="Frequency of existing variant in NHLBI-ESP "
         "European American population",
     ),
     "gnomAD_AF": AnnotationEntry(
         "gnomAD_AF",
-        float,
+        float32,
         description="Frequency of existing variant in gnomAD exomes "
         "combined population",
     ),
     "gnomAD_AFR_AF": AnnotationEntry(
         "gnomAD_AFR_AF",
-        float,
+        float32,
         description="Frequency of existing variant in gnomAD exomes "
         "African/American population",
     ),
     "gnomAD_AMR_AF": AnnotationEntry(
         "gnomAD_AMR_AF",
-        float,
+        float32,
         description="Frequency of existing variant in gnomAD exomes "
         "American population",
     ),
     "gnomAD_ASJ_AF": AnnotationEntry(
         "gnomAD_ASJ_AF",
-        float,
+        float32,
         description="Frequency of existing variant in gnomAD exomes "
         "Ashkenazi Jewish population",
     ),
     "gnomAD_EAS_AF": AnnotationEntry(
         "gnomAD_EAS_AF",
-        float,
+        float32,
         description="Frequency of existing variant in gnomAD exomes "
         "East Asian population",
     ),
     "gnomAD_FIN_AF": AnnotationEntry(
         "gnomAD_FIN_AF",
-        float,
+        float32,
         description="Frequency of existing variant in gnomAD exomes "
         "Finnish population",
     ),
     "gnomAD_NFE_AF": AnnotationEntry(
         "gnomAD_NFE_AF",
-        float,
+        float32,
         description="Frequency of existing variant in gnomAD exomes "
         "Non-Finnish European population",
     ),
     "gnomAD_OTH_AF": AnnotationEntry(
         "gnomAD_OTH_AF",
-        float,
+        float32,
         description="Frequency of existing variant in gnomAD exomes "
         "combined other combined populations",
     ),
     "gnomAD_SAS_AF": AnnotationEntry(
         "gnomAD_SAS_AF",
-        float,
+        float32,
         description="Frequency of existing variant in gnomAD exomes "
         "South Asian population",
     ),
     "MAX_AF": AnnotationEntry(
         "MAX_AF",
-        float,
+        float32,
         description="Maximum observed allele frequency in "
         "1000 Genomes, ESP and gnomAD",
     ),
-    # "MAX_AF_POPS": AnnotationEntry(
-    #     "MAX_AF_POPS",
-    #     description="Populations in which maximum allele frequency was observed",
-    # ),
+    "MAX_AF_POPS": AnnotationListEntry(
+        "MAX_AF_POPS",
+        sep="&",
+        description="Populations in which maximum allele frequency was observed",
+    ),
     "CLIN_SIG": AnnotationListEntry(
         "CLIN_SIG",
         "&",
@@ -464,32 +630,35 @@ KNOWN_ANN_TYPE_MAP_VEP = {
         "NB: not available for GRCh37",
     ),
     "TSL": AnnotationEntry(
-        "TSL", description="Transcript support level. " "NB: not available for GRCh37"
+        "TSL", description="Transcript support level. NB: not available for GRCh37"
     ),
-    "PUBMED": AnnotationEntry(
-        "PUBMED", description="Pubmed ID(s) of publications that cite existing variant"
+    "PUBMED": AnnotationListEntry(
+        "PUBMED",
+        description="Pubmed ID(s) of publications that cite existing variant",
+        sep="&",
+    ),
+    "SOMATIC": AnnotationListEntry(
+        "SOMATIC",
+        description="Somatic status of existing variant(s); "
+        "multiple values correspond to multiple values "
+        "in the Existing_variation field",
+        sep="&",
+    ),
+    "PHENO": AnnotationListEntry(
+        "PHENO",
+        sep="&",
+        description="Indicates if existing variant is associated with a phenotype, "
+        "disease or trait; "
+        "multiple values correspond to multiple values "
+        "in the Existing_variation field",
+    ),
+    # TODO list type with unconfirmed sep
+    "GENE_PHENO": AnnotationListEntry(
+        "GENE_PHENO",
+        sep="&",
+        description="Indicates if overlapped gene is associated with a phenotype, "
+        "disease or trait",
     ),
-    # TODO list type with unknown sep
-    # "SOMATIC": AnnotationEntry(
-    #     "SOMATIC",
-    #     description="Somatic status of existing variant(s); "
-    #                 "multiple values correspond to multiple values "
-    #                 "in the Existing_variation field",
-    # ),
-    # TODO list type with unknown sep
-    # "PHENO": AnnotationEntry(
-    #     "PHENO",
-    #     description="Indicates if existing variant is associated with a phenotype, "
-    #                 "disease or trait; "
-    #                 "multiple values correspond to multiple values "
-    #                 "in the Existing_variation field",
-    # ),
-    # TODO list type with unknown sep
-    # "GENE_PHENO": AnnotationEntry(
-    #     "GENE_PHENO",
-    #     description="Indicates if overlapped gene is associated with a phenotype, "
-    #                 "disease or trait",
-    # ),
     "ALLELE_NUM": AnnotationEntry(
         "ALLELE_NUM",
         int,
@@ -534,7 +703,7 @@ KNOWN_ANN_TYPE_MAP_VEP = {
     ),
     "OverlapPC": AnnotationEntry(
         "OverlapPC",
-        float,
+        float32,
         description="Percentage of corresponding structural variation feature "
         "overlapped by the given input",
     ),
@@ -546,6 +715,64 @@ KNOWN_ANN_TYPE_MAP_VEP = {
     "AMBIGUITY": AnnotationEntry(
         "AMBIGUITY", description="IUPAC allele ambiguity code"
     ),
+    "Amino_acids": AnnotationListEntry(
+        "Amino_acids", description="Reference and variant amino acids", sep="/"
+    ),
+    "Codons": AnnotationListEntry(
+        "Codons", description="Reference and variant codon sequence", sep="/"
+    ),
+    # TODO HGNC_ID description
+    "HGNC_ID": AnnotationEntry("HGNC_ID"),
+    "MANE": AnnotationEntry(
+        "MANE", description="Matched Annotation from NCBI and EMBL-EBI (MANE)."
+    ),
+    "MANE_SELECT": AnnotationEntry(
+        "MANE_SELECT",
+        description="Matched Annotation from NCBI and EMBL-EBI (MANE) "
+        "canonical transcript, indicated by the respective RefSeq NM ID."
+        "For more info, see: "
+        "https://www.ensembl.org/info/genome/genebuild/mane.html",
+    ),
+    "MANE_PLUS_CLINICAL": AnnotationEntry(
+        "MANE_PLUS_CLINICAL",
+        description="Matched Annotation from NCBI and EMBL-EBI (MANE) "
+        "transcripts beyond MANE_SELECT, that are important clinically. "
+        "Indicated by their RefSeq NM IDs."
+        "For more info, see: "
+        "https://www.ensembl.org/info/genome/genebuild/mane.html",
+    ),
+    "GO": AnnotationEntry("GO", description="Gene ontology (GO) terms."),
+    "miRNA": AnnotationEntry(
+        "miRNA",
+        description="Determines where in the secondary structure of a miRNA "
+        "a variant falls",
+    ),
+    "Existing_variation": AnnotationListEntry(
+        "Existing_variation",
+        sep="&",
+        description="Identifier(s) of co-located known variants",
+    ),
+    "LoFtool": AnnotationEntry(
+        "LoFtool",
+        typefunc=float32,
+        description="Provides a rank of genic intolerance and "
+        "consequent susceptibility to disease based on the ratio of Loss-of-function "
+        "(LoF) to synonymous mutations.",
+    ),
+    "REVEL": AnnotationEntry(
+        "REVEL",
+        typefunc=float32,
+        description="Estimate of the pathogenicity of missense variants. "
+        "Please cite the REVEL publication alongside the VEP if you use this "
+        "resource: https://www.ncbi.nlm.nih.gov/pubmed/27666373",
+    ),
+    "ExACpLI": AnnotationEntry(
+        "ExACpLI",
+        typefunc=float32,
+        description="Probabililty of a gene being loss-of-function intolerant (pLI). "
+        "Please cite the respective ExAC publication alongside the VEP if you use this "
+        "resource: https://www.ncbi.nlm.nih.gov/pubmed/27535533",
+    ),
 }
 
 KNOWN_ANN_TYPE_MAP = {**KNOWN_ANN_TYPE_MAP_SNPEFF, **KNOWN_ANN_TYPE_MAP_VEP}
